import json
from pathlib import Path

from hatchling.metadata.plugin.interface import MetadataHookInterface

class CustomMetadataHook(MetadataHookInterface):
    def update(self, metadata):
        metadata_file = Path(self.root) / "metadata.json"
        if not metadata_file.exists():
            raise FileNotFoundError(f"Metadata file not found at: {metadata_file}")

        with metadata_file.open(encoding="utf-8") as f:
            meta = json.load(f)

        required_fields = ["version", "description", "authors"]
        for field in required_fields:
            if field not in meta:
                raise ValueError(f"Missing required field '{field}' in metadata.json")

        metadata["version"] = meta["version"]
        metadata["description"] = meta["description"]
        metadata["authors"] = meta["authors"]



import datetime
import inspect
import json
import os
import pathlib
import time
import warnings

_DELAY = 604800 # 1 week


def _get_tdx_path(endpoint: str | pathlib.Path) -> pathlib.Path:
    return pathlib.Path(os.path.join(os.path.expanduser("~"), ".TyraDex")) / endpoint


class TDXFile:
    @classmethod
    def create(cls, endpoint, data):
        path = _get_tdx_path(endpoint + '.tdx')
        if path.exists():
            os.remove(path)
        tdx = cls(path)
        tdx._data = data
        tdx._limit = int(time.time()) + _DELAY
        tdx.save()
        return tdx

    @classmethod
    def get(cls, endpoint):
        path = _get_tdx_path(endpoint)
        return cls(path)

    def __init__(self, path: str | pathlib.Path):
        self._path = pathlib.Path(str(path) + ('' if str(path).endswith('.tdx') else '.tdx'))
        if not self._path.parent.exists():
            self._path.parent.mkdir(parents=True, exist_ok=True)

        self._endpoint: str = str(self._path).removeprefix(str(_get_tdx_path('')))[1:].removesuffix('.tdx')
        self._limit: int = 0
        self._data = {}

        self.load()

    def __repr__(self):
        return (
            "<data:{data}, limit: {limit}, endpoint: {endpoint}>"
            .format(
                data=str(self._data).replace('\n', ' '),
                limit=self._limit,
                endpoint=self._endpoint
            )
        )

    def __str__(self):
        return json.dumps(self._data, ensure_ascii=False, indent=4)

    @property
    def exists(self):
        return self._path.exists()

    @property
    def is_valid(self):
        return self._limit > time.time()

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def limit(self):
        return datetime.datetime.fromtimestamp(int(self._limit))

    @property
    def data(self):
        if self._limit < time.time():
            stacks = [s.function for s in inspect.stack()]
            if 'create' not in stacks:
                sl = 2
                if stacks[1] == "save":
                    sl = 3
                    if stacks[2] == '__init__':
                        sl = 4
                warnings.warn('Data is out of date.', UserWarning, stacklevel=sl)
        return self._data

    def update(self, data):
        if isinstance(data, dict) or isinstance(data, list):
            self._data = data
            self._limit = int(time.time()) + _DELAY
            self.save()
        else:
            try:
                self.update(list(iter(data)))
            except TypeError:
                raise TypeError("Data must be iterable.")

    def load(self):
        if self._path.exists():
            with open(self._path, "r", encoding='utf-8') as f:
                data = ""
                for line in f.readlines():
                    if line.startswith("$") or line.isspace() or line == "\n":
                        pass
                    else:
                        data += line

                try:
                    self._endpoint = data.splitlines()[0]
                except TypeError:
                    raise TypeError("Critical error, the endpoint is bad.")

                try:
                    self._limit = int(data.splitlines()[1])
                    if self._limit < time.time():
                        if inspect.stack()[1].function != "__init__":
                            warnings.warn('The data is out of date.', UserWarning, stacklevel=2)
                except (ValueError, TypeError):
                    warnings.warn('Limit must be an int. The data will be considered outdated.', RuntimeWarning, stacklevel=2)
                    self._limit = 0

                try:
                    self._data = json.loads('\n'.join(data.splitlines()[2:]))
                except (json.decoder.JSONDecodeError, TypeError):
                    self._limit = 0
                    self._data = {}

    def save(self):
        with open(self._path, "w", encoding='utf-8') as f:
            f.write(
                "{endpoint}\n{limit}\n{data}"
                .format(
                    endpoint=self._endpoint,
                    limit=self._limit,
                    data=json.dumps(self.data, ensure_ascii=False, indent=4)
                )
            )


import json
import os
import sys

try:
    import API as API
except ModuleNotFoundError:
    import Tyradex.API as API

def help_message():
    print("Usage:")
    print('py' if os.name == 'nt' else 'python3', end='')
    print(" -m Tyradex [options] <endpoint>\n")
    print("Description:")
    print("\tTyradex for Python is an easy-to-use API based on the web version of Tyradex by Yarkis and Ashzuu.\n")
    print("Options:")
    print("\t-h, --help     Display this help message and exit.")
    print("\t-v, --version  Display the version and exit.")

if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1].lower() in ['-h', '--help']:
        help_message()
    elif len(sys.argv) > 1 and sys.argv[1].lower() in ['-v', '--version']:
        with open('../metadata.json', 'r') as f:
            print(API.VERSION)
    else:
        print(json.dumps(API.Tyradex.call('/'.join(sys.argv[1:])), indent=4))


"""
ABC module

All schema for Tyradex.
"""
import locale

DEFAULT_IMG = "https://raw.githubusercontent.com/LassaInora/Tyradex/20545b1b15d72564987c8dcebc39501e91ca5f40/assets/img/clone.png"

class NameModel:
    """ A multilingual name

    Attributes:
        fr (str): French name.
        en (str): English name.
        jp (str): Japan name.
    """
    def __init__(self, **data):
        """ A multilingual name.

        Args:
            fr (str): French name.
            en (str): English name.
            jp (str): Japan name.
        """
        self._fr = data.get('fr')
        self._en = data.get('en')
        self._jp = data.get('jp')

    def __str__(self):
        match locale.getlocale()[0][:2]:
            case "fr":
                return ', '.join(self._fr) if isinstance(self._fr, list) else str(self._fr)
            case "ja":
                return ', '.join(self._jp) if isinstance(self._jp, list) else str(self._jp)
            case _:
                return ', '.join(self._en) if isinstance(self._en, list) else str(self._en)

    @property
    def fr(self):
        """ The French name.

        Returns:
            str: French name.
        """
        return self._fr

    @property
    def en(self):
        """ The English name.

        Returns:
            str: English name.
        """
        return self._en

    @property
    def jp(self):
        """ The Japan name.

        Returns:
            str: Japan name.
        """
        return self._jp

class SimpleSpriteModel:
    """ A simple sprite.

    Attributes:
        regular (str): Regular sprite.
        shiny (str): Shiny sprite.

    """

    def __init__(self, **data):
        """ A simple sprite.

        Args:
            regular (str): Regular sprite.
            shiny (str): Shiny sprite.
        """
        self._regular = sprite if (sprite := data.get('regular')) else DEFAULT_IMG
        self._shiny = sprite if (sprite := data.get('shiny')) else DEFAULT_IMG

    @property
    def regular(self):
        """ The regular sprite.

        Returns:
            str: Regular sprite.
        """
        return self._regular

    @property
    def shiny(self):
        """ The shiny sprite.

        Returns:
            str: Shiny sprite.
        """
        return self._shiny

class DetailedSpriteModel:
    """ A detailed sprite.

    Attributes:
        regular (str): Regular sprite.
        shiny (str): Shiny sprite.
        gmax (SimpleSpriteModel): GigaMax sprites.
    """

    def __init__(self, **data):
        """ A detailed sprite model.

        Args:
            regular (str): Regular sprite.
            shiny (str): Shiny sprite.
            gmax (dict): GigaMax sprites.
        """
        self._regular = sprite if (sprite := data.get('regular')) else DEFAULT_IMG
        self._shiny = sprite if (sprite := data.get('shiny')) else DEFAULT_IMG
        self._gmax = SimpleSpriteModel(**(gmax if (gmax := data.get('gmax')) else {}))

    @property
    def regular(self):
        """ The regular sprite.

        Returns:
            str: Regular sprite.
        """
        return self._regular

    @property
    def shiny(self):
        """ The shiny sprite.

        Returns:
            str: Shiny sprite.
        """
        return self._shiny

    @property
    def gmax(self):
        """ The GigaMax sprite.

        Returns:
            str: GigaMax sprite.
        """
        return self._gmax

class TypeModel:
    """ A type

    Attributes:
        name (str): Name.
        image (str): Image.
    """

    def __init__(self, **data):
        """ A type.

        Args:
            name (str): Name.
            image (str): Image.
        """
        self._name = str(data.get('name'))
        self._image = sprite if (sprite := data.get('image')) else DEFAULT_IMG

    @property
    def name(self):
        """ Name.

        Returns:
            str: Name.
        """
        return self._name

    @property
    def image(self):
        """ Image.

        Returns:
            str: Image.
        """
        return self._image

class TalentModel:
    """ A talent

    Attributes:
        name (str): Name.
        tc (bool): Is hidden talent.
    """

    def __init__(self, **data):
        """ A talent

        Args:
            name (str): Name.
            tc (bool): Is hidden talent.
        """
        self._name = str(data.get('name'))
        self._tc = tc if (tc := data.get('tc')) else False

    @property
    def name(self):
        """ Name.

        Returns:
            str: Name.
        """
        return self._name

    @property
    def tc(self):
        """ Is hidden talent.

        Returns:
            bool: Is hidden talent.
        """
        return self._tc

class StatModel:
    """ A stats.

    Attributes:
        hp (int): The health points.
        atk (int): The attack points.
        def (int): The defense points.
        spe_atk (int): The special attack points.
        spe_def (int): The special defense points.
        vit (int): The vitesse points.
    """

    def __init__(self, **data):
        """ A stats.

        Args:
            hp (int): The health points.
            atk (int): The attack points.
            def (int): The defense points.
            spe_atk (int): The special attack points.
            spe_def (int): The special defense points.
            vit (int): The vitesse points.
        """
        self._hp = hp_ if (hp_ := data.get('hp')) else 0
        self._atk = atk_ if (atk_ := data.get('atk')) else 0
        self._def = def_ if (def_ := data.get('def')) else 0
        self._spe_atk = spe_atk_ if (spe_atk_ := data.get('spe_atk')) else 0
        self._spe_def = spe_def_ if (spe_def_ := data.get('spe_def')) else 0
        self._vit = vit_ if (vit_ := data.get('vit')) else 0

    @property
    def hp(self):
        """ The health points.

        Returns:
            int: health points
        """
        return self._hp

    @property
    def atk(self):
        """ The attack points.

        Returns:
            int: attack points
        """
        return self._atk

    @property
    def def_(self):
        """ The defense points.

        Returns:
            int: defense points
        """
        return self._def

    @property
    def spe_atk(self):
        """ The special attack points.

        Returns:
            int: special attack points
        """
        return self._spe_atk

    @property
    def spe_def(self):
        """ The special defense points.

        Returns:
            int: special defense points
        """
        return self._spe_def

    @property
    def vit(self):
        """ The vitesse points.

        Returns:
            int: vitesse points
        """
        return self._vit

class ResistanceModel:
    """ Resistance.

    Attributes:
        name (str): Name.
        multiplier (float): Multiplier.
    """

    def __init__(self, **data):
        """ Resistance.

        Args:
            name (str): Name.
            multiplier (float): Multiplier.
        """
        self._name = str(data.get('name'))
        self._multiplier = multiplier if (multiplier := data.get('multiplier')) else 1

    @property
    def name(self):
        """ Name.

        Returns:
            str: Name.
        """
        return self._name

    @property
    def multiplier(self):
        """ Multiplier.

        Returns:
            float: Multiplier.
        """
        return self._multiplier

class PokemonEvolutionModel:
    """ Pokémon evolution.

    Attributes:
        pokedex_id (int): Pokédex ID.
        name (str): Name.
        condition (str): Conditions.
    """

    def __init__(self, **data):
        """ Pokémon evolution.

        Args:
            pokedex_id (int): Pokédex ID.
            name (str): Name.
            condition (str): Conditions.
        """
        self._pokedex_id = pokedex_id if (pokedex_id := data.get('pokedex_id')) else -1
        self._name = str(data.get('name'))
        self._condition = str(data.get('condition'))

    @property
    def pokedex_id(self):
        """ Pokédex ID.

        Returns:
            int: Pokédex ID.
        """
        return self._pokedex_id

    @property
    def name(self):
        """ Name.

        Returns:
            str: Name.
        """
        return self._name

    @property
    def condition(self):
        """ Condition.

        Returns:
            str: Condition.
        """
        return self._condition

class MegaEvolutionModel:
    """ Mega evolution.

    Attributes:
        name (str): Names.
        orbes (list[str]): Orbes.
        sprites (SimpleSpriteModel): Sprites.
    """

    def __init__(self, **data):
        """ Mega evolution.

        Args:
            name (str): Names.
            orbes (list[str]): Orbes.
            sprites (dict): Sprites.
        """
        self._name = str(data.get('name'))
        self._orbes = orbes if (orbes := data.get('orbe')) else []
        self._sprites = SimpleSpriteModel(**(sprites if (sprites := data.get('sprites')) else {}))

    @property
    def name(self):
        """ Name.

        Returns:
            str: Name.
        """
        return self._name

    @property
    def orbes(self):
        """ Orbes.

        Returns:
            list[str]: Orbes.
        """
        return self._orbes

    @property
    def sprites(self):
        """ Sprites.

        Returns:
            SimpleSpriteModel: Sprites.
        """
        return self._sprites

class EvolutionModel:
    """ Evolution.

    Attributes:
        pre (list[PokemonEvolutionModel]): Pre-Evolutions.
        next (list[PokemonEvolutionModel]): Next-Evolutions.
        mega (list[MegaEvolutionModel]): Mega-Evolutions.
    """

    def __init__(self, **data):
        """ Evolution.

        Args:
            pre (list[dict]): Pre-Evolutions.
            next (list[dict]): Next-Evolutions.
            mega (list[dict]): Mega-Evolutions.
        """
        pres = data.get('pre')
        self._pre = [PokemonEvolutionModel(**pre) for pre in (pres if pres else [])]
        nexts = data.get('next')
        self._next = [PokemonEvolutionModel(**next) for next in (nexts if nexts else [])]
        megas = data.get('mega')
        self._mega = [MegaEvolutionModel(**mega) for mega in (megas if megas else [])]

    @property
    def pre(self):
        """ Pre-Evolutions.

        Returns:
            list[PokemonEvolutionModel]: Pre-Evolutions.
        """
        return self._pre

    @property
    def next(self):
        """ Next-Evolutions.

        Returns:
            list[PokemonEvolutionModel]: Next-Evolutions.
        """
        return self._next

    @property
    def mega(self):
        """ Mega-Evolutions.

        Returns:
            list[PokemonEvolutionModel]: Mega-Evolutions.
        """
        return self._mega

class SexeModel:
    """ Sexe.

    Attributes:
        male (float): Pourcentage of male.
        female (float): Pourcentage of female.
    """

    def __init__(self, **data):
        """ Sexe.

        Args:
            male (float): Pourcentage of male.
            female (float): Pourcentage of female.
        """
        self._male = male if (male := data.get('male')) else 0.0
        self._female = female if (female := data.get('female')) else 0.0

    @property
    def male(self):
        """ Pourcentage of male.

        Returns:
            float: Pourcentage of male.
        """
        return self._male

    @property
    def female(self):
        """ Pourcentage of female.

        Returns:
            float: Pourcentage of female.
        """
        return self._female

class RegionalFormModel:
    """ Regional form

    Attributes:
        region (str): Region.
        name (NameModel): Name.
    """

    def __init__(self, **data):
        """ Regional form

        Args:
            region (str): Region.
            name (dict): Name.
        """
        self._region = str(data.get('region'))
        self._name = NameModel(**(name if (name := data.get('name')) else {}))

    @property
    def region(self):
        """ Region.

        Returns:
            str: Region.
        """
        return self._region

    @property
    def name(self):
        """ Name.

        Returns:
            NameModel: Name.
        """
        return self._name


try:
    import API as API
except ModuleNotFoundError:
    import Tyradex.API as API
try:
    import abc as abc
except ModuleNotFoundError:
    import Tyradex.abc as abc

__version__ = API.VERSION

class Pokemon:
    """ A Pokémon

    Attributes:
        pokedex_id (int): Pokédex ID
        generation (int): Generation
        category (str): Category
        name (abc.NameModel): Names
        sprites (abc.DetailedSpriteModel): Sprites
        types (list[abc.TypeModel]): Types
        talents (list[abc.TalentModel]): Talents
        stats (abc.StatModel): Stats
        resistances (list[abc.ResistanceModel]): Resistances
        evolutions (abc.EvolutionModel): Evolutions
        height (float): Height
        weight (float): Weight
        egg_groups (list[str]): Egg groups
        sexe (abc.SexeModel): Sexe
        catch_rate (int): Catch rate
        level_100 (int): Level 100
        formes (list[abc.RegionalFormModel]): Formes
    """

    @classmethod
    def all(cls):
        """ Get the list of different Pokémon.

        Returns:
            list[Pokemon]: The list of all Pokémon
        """
        response = API.Tyradex.call('pokemon')
        return [cls(**pokemon_data) for pokemon_data in (response if response else [])]

    @classmethod
    def get(cls, pokemon_id, region=..., talent=...):
        """ Get a Pokémon's information.

        Args:
            pokemon_id (str | int): The Pokémon's ID in the National Pokédex or its name.
            region (str): The Pokémon region. (Allows you to retrieve information on a Pokémon's regional form.)
            talent (str): The Pokémon's talent. (Allows you to recover the Pokémon's resistances based on its ability.)

        Returns:
            Pokemon: The Pokémon's information
        """
        endpoint = 'pokemon/{id_}'.format(id_=pokemon_id)
        if region is not ...:
            endpoint += '/{region}'.format(region=region)
        if talent is not ...:
            endpoint += '?talent={talent}'.format(talent=talent)
        return cls(**response) if (response := API.Tyradex.call(endpoint)) else None

    def __init__(self, **data):
        """ A Pokémon

        Args:
            pokedex_id (int): Pokédex ID
            generation (int): Generation
            category (str): Category
            name (dict): Names
            sprites (dict): Sprites
            types (list[dict]): Types
            talents (list[dict]): Talents
            stats (dict): Stats
            resistances (list[dict]): Resistances
            evolutions (dict): Evolutions
            height (str): Height
            weight (str): Weight
            egg_groups (list[str]): Egg groups
            sexe (dict): Sexe
            catch_rate (int): Catch rate
            level_100 (int): Level 100
            formes (list[dict]): Pokémon formes
        """
        self._pokedex_id = (pokedex_id if (pokedex_id := data.get('pokedex_id')) else -1)
        self._generation = (generation if (generation := data.get('generation')) else -1)
        self._category = str(data.get('category'))
        self._name = abc.NameModel(**data.get('name'))
        self._sprites = abc.DetailedSpriteModel(**(sprites if (sprites := data.get('sprites')) else {}))
        types = data.get('types')
        self._types = [abc.TypeModel(**type_data) for type_data in (types if types else [])]
        talents = data.get('talents')
        self._talents = [abc.TalentModel(**talent_data) for talent_data in (talents if talents else [])]
        self._stats = abc.StatModel(**(stats if (stats := data.get('stats')) else {}))
        resistances = data.get('resistances')
        self._resistances = [abc.ResistanceModel(**resistance_data) for resistance_data in (resistances if resistances else [])]
        self._evolutions = abc.EvolutionModel(**(evolutions if (evolutions := data.get('evolutions')) else {}))
        self._height = height.split(' ')[0].replace(',', '.') if (height := data.get('height')) else 0.0
        self._weight = weight.split(' ')[0].replace(',', '.') if (weight := data.get('weight')) else 0.0
        self._egg_groups = egg_groups if (egg_groups := data.get('egg_groups')) else []
        self._sexe = abc.SexeModel(**(sexe if (sexe := data.get('sexe')) else {}))
        self._catch_rate = catch_rate if (catch_rate := data.get('catch_rate')) else 0
        self._level_100 = level_100 if (level_100 := data.get('level_100')) else -1
        formes = data.get('formes')
        self._formes = [abc.RegionalFormModel(**forme_data) for forme_data in (formes if formes else [])]

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return '<{id_}:{name}>'.format(id_=self._pokedex_id, name=self.name.en)

    @property
    def pokedex_id(self):
        """ Pokédex ID.

        Returns:
            int: Pokédex ID.
        """
        return self._pokedex_id

    @property
    def generation(self):
        """ Generation.
        
        Returns:
            int: Génération.
        """
        return self._generation
    
    @property
    def category(self):
        """ Category.

        Returns:
            str: Category.
        """
        return self._category

    @property
    def name(self):
        """ Name.
        
        Returns:
            abc.NameModel: Names.
        """
        return self._name
    
    @property
    def sprites(self):
        """ Sprites.

        Returns:
            DetailedSpriteModel : Sprites.
        """
        return self._sprites

    @property
    def types(self):
        """ Types.
        
        Returns:
            list[abc.TypeModel]: Types.
        """
        return self._types
    
    @property
    def talents(self):
        """ Talents.

        Returns:
            list[abc.TalentModel]: Talents.
        """
        return self._talents

    @property
    def stats(self):
        """ Stats
        
        Returns:
            abc.StatModel: Stats.
        """
        return self._stats
    
    @property
    def resistances(self):
        """ Resistances.

        Returns:
            list[abc.ResistanceModel]: Resistances.
        """
        return self._resistances

    @property
    def evolutions(self):
        """ Evolutions.
        
        Returns:
            
        """
        return self._evolutions
    
    @property
    def height(self):
        """ Height.

        Returns:
            float: Pokémon height
        """
        return self._height

    @property
    def weight(self):
        """ Weight
        
        Returns:
            float: Pokémon weight
        """
        return self._weight
    
    @property
    def egg_groups(self):
        """ Egg groups

        Returns:

        """
        return self._egg_groups

    @property
    def sexe(self):
        """ Sexe
        
        Returns:
            
        """
        return self._sexe
    
    @property
    def catch_rate(self):
        """ Catch rate

        Returns:

        """
        return self._catch_rate

    @property
    def level_100(self):
        """ Level 100
        
        Returns:
            
        """
        return self._level_100
    
    @property
    def formes(self):
        """ Formes

        Returns:

        """
        return self._formes

class Generation:
    """ Generation.

    Attributes:
        generation (int): Generation.
        from (int): Which index starts.
        to_ (int): Which index ended.
    """

    @classmethod
    def all(cls):
        """ Get the list of different generations

        Returns:
            list[Generation]: The list of all generation
        """
        response = API.Tyradex.call('gen')
        return [cls(**gen_data) for gen_data in (response if response else [])]

    @classmethod
    def get(cls, generation):
        """ Get the list of Pokémon of a generation.

        Args:
            generation (int): The generation number.

        Returns:
            list[Pokemon]: The list of Pokémon of a generation.
        """
        response = API.Tyradex.call('gen/{id_}'.format(id_=generation))
        return [Pokemon(**pokemon_data) for pokemon_data in (response if response else [])]

    def __init__(self, **data):
        """ Generation.

        Args:
            generation (int): Generation.
            from_ (int): Which index starts.
            to_ (int): Which index ended.
        """
        self._generation = generation if (generation := data.get('generation')) else -1
        self._from = from_ if (from_ := data.get('from')) else -1
        self._to = to if (to := data.get('to')) else -1

    def __str__(self):
        return 'Gen {gen}'.format(gen=self._generation)

    def __repr__(self):
        return '<gen {g} from {f} to {t}>'.format(g=self._generation, f=self._from, t=self._to)

    @property
    def generation(self):
        """ Generation.

        Returns:
            int: Generation.
        """
        return self._generation

    @property
    def from_(self):
        """ Which index starts.

        Returns:
            int: Which index started.
        """
        return self._from

    @property
    def to_(self):
        """ Which index ended.

        Returns:
            int: Which index ended
        """
        return self._to

class Type:
    """ Type.

    Attributes:
        type_id (int): Type ID.
        name (abc.NameModel): Names.
        sprites (string): Sprites.
        resistances (list[abc.ResistanceModel]): Resistances.
        pokemons (list[Pokemon]): All Pokémon of this type.
    """

    @classmethod
    def all(cls):
        """ Get the list of all types.

        Returns:
            list[Type]: The list of all types.
        """
        response = API.Tyradex.call('types')
        return [cls(**type_data) for type_data in (response if response else [])]

    @classmethod
    def get(cls, type_1, type_2=...):
        """ Get the list of Pokémon of a generation.

        Args:
            type_1 (int | str): The First type identifier, or its English or French name.
            type_2 (int | str): The Second type identifier, or its English or French name.

        Returns:
            Type: The Type
        """
        if type_1 == type_2:
            type_2 = ...
        elif type_2 is not ...:
            type_1, type_2 = sorted((type_1, type_2))
        endpoint = 'types/{type_1}'.format(type_1=type_1)
        if type_2 is not ...:
            endpoint += '/{type_2}'.format(type_2=type_2)
        return Type(**(type_ if (type_ := API.Tyradex.call(endpoint)) else {}))

    def __init__(self, **data):
        """ Type.

        Attributes:
            type_id (int): Type ID.
            name (dict): Names.
            sprites (string): Sprites.
            resistances (list[dict]): Resistances.
            pokemons (list[dict]): All Pokémon of this type.
        """
        self._type_id = id_ if (id_ := data.get('type_id')) else -1
        self._name = abc.NameModel(**(names if (names := data.get('name')) else {}))
        self._sprites = sprites if (sprites := data.get('sprites')) else abc.DEFAULT_IMG
        resistances = data.get('resistances')
        self._resistances = [abc.ResistanceModel(**resistance_data) for resistance_data in (resistances if resistances else [])]
        pokemon = data.get('pokemons')
        self._pokemons = [Pokemon(**pokemon_data) for pokemon_data in (pokemon if pokemon else [])]

    def __str__(self):
        return str(self._name)

    def __repr__(self):
        return '<{id_}:{name}>'.format(id_=self._type_id, name=self.name.en)

    @property
    def type_id(self):
        """ Type ID.

        Returns:
            int: Type ID.
        """
        return self._type_id

    @property
    def name(self):
        """ Names.

        Returns:
            abc.NameModel: Names.
        """
        return self._name

    @property
    def sprites(self):
        """ Sprites.

        Returns:
            str: Sprites.
        """
        return self._sprites

    @property
    def resistances(self):
        """ Resistances.

        Returns:
            list[abc.ResistanceModel]: Resistances.
        """
        return self._resistances

    @property
    def pokemons(self):
        """ Pokémon.

        Returns:
            list[Pokemon]: Pokémon.
        """
        return self._pokemons



if __name__ == '__main__':
    Type.all()



import pratik.text


class TyradexError(Exception):
    def __init__(self, *args):
        super().__init__(*args)

    def __str__(self):
        return str(self.__class__.__name__) + ' : ' + ', '.join(self.args)

    def print(self):
        print(pratik.text.Color.RED + str(self) + pratik.text.Color.STOP)

class PageNotFound(TyradexError):
    def __init__(self, endpoint=''):
        super().__init__(
            "The page{endpoint} was not found"
            .format(
                endpoint=(' "' + endpoint + '"') if endpoint != '' else ''
            )
        )


import ctypes
import json
import os
import pathlib
import platform
import subprocess

import requests

try:
    import TyradexErrors as TyradexErrors
except ModuleNotFoundError:
    import Tyradex.TyradexErrors as TyradexErrors
try:
    import tdx_file as TDXFile
except ModuleNotFoundError:
    import Tyradex.tdx_file as TDXFile

with open('../metadata.json', 'r') as f:
    _METADATA = json.load(f)
    VERSION = 'Tyradex for Python version {version} on {os} {os_release} ({os_version})'.format(
        version=_METADATA['version'],
        os=platform.system(),
        os_release=platform.release(),
        os_version=platform.version(),
    )

_CACHE_FOLDER = pathlib.Path(os.path.join(os.path.expanduser("~"), ".TyraDex"))
if not os.path.exists(_CACHE_FOLDER):
    os.makedirs(_CACHE_FOLDER)
    if platform.system() == "Windows":  # Windows
        ctypes.windll.kernel32.SetFileAttributesW(_CACHE_FOLDER, 0x02)
    elif platform.system() == "Darwin":  # macOS
        subprocess.run(["chflags", "hidden", _CACHE_FOLDER])


class Tyradex:
    _URL = "https://tyradex.app/api/v1/"

    _HEADER = {
        "User-Agent": "Tyradex For Python Version {version}".format(version=_METADATA["version"]),
        "From": "https://github.com/LassaInora/TyraDex",
        'Content-type': 'application/json'
    }

    @classmethod
    def call(cls, endpoint):
        """
                Sends a request to the Tyradex API with local cache handling.

                If the requested endpoint data is available and valid in the cache,
                it is returned directly without performing an HTTP request.
                Otherwise, a GET request is made to the API. On success, the data is
                cached and returned. If an error occurs (e.g., 404), a custom error
                message is displayed and `None` is returned.

                Args:
                    endpoint (str): The API endpoint to query (e.g., "pokemon/1").

                Returns:
                    dict | None: The data returned as a dictionary, or `None` if an
                    error occurred (e.g., page not found).
                """
        cache = TDXFile.get(endpoint)
        if not cache.exists or not cache.is_valid:
            request = requests.get(cls._URL + endpoint, headers=cls._HEADER)
            try:
                rjson = request.json()
            except requests.exceptions.JSONDecodeError:
                rjson = {
                    "status": request.status_code,
                    "message": request.reason
                }
            if isinstance(rjson, dict) and rjson.get('status') == 404:
                src.TyradexErrors.PageNotFound(endpoint).print()
                return None
            else:
                cache.update(request.json())
        return cache.data
